"""
Authors:
        ITryagain <long452a@163.com>

Reference:
        https://www.ibm.com/developerworks/community/blogs/jfp/entry/Fast_Computation_of_AUC_ROC_score?lang=en
        https://www.kaggle.com/uberkinder/efficient-metric
        https://www.kaggle.com/artgor
introduce:
        this file contains the use of models such as LightGBM, XGBoost, CatBoost and so on
"""
import time
import gc
import numpy as np
import pandas as pd
import lightgbm as lgb
import xgboost as xgb
from catboost import CatBoostRegressor, CatBoostClassifier
from sklearn import metrics
from numba import jit
from preprocessing.Encoding import BetaEncoder
from tqdm import tqdm


# compute auc faster than sklearn
# However, when there are ties in predictions computed value may differ from sklearn value.
# https://www.ibm.com/developerworks/community/blogs/jfp/entry/Fast_Computation_of_AUC_ROC_score?lang=en
@jit
def fast_auc(y_true, y_prob):
    y_true = np.asarray(y_true)
    y_true = y_true[np.argsort(y_prob)]
    nfalse = 0
    auc = 0
    n = len(y_true)
    for i in range(n):
        y_i = y_true[i]
        nfalse += (1 - y_i)
        auc += y_i * nfalse
    auc /= (nfalse * (n - nfalse))
    return auc


# define metric by ourself
def eval_auc(preds, dtrain):
    return 'auc', fast_auc(dtrain, preds), True


def group_mean_log_mae(y_true, y_pred, group, floor=1e-9):
    """
    Fast metric computation for this competition: https://www.kaggle.com/c/champs-scalar-coupling
    Code is from this kernel: https://www.kaggle.com/uberkinder/efficient-metric
    """
    maes = (y_true - y_pred).abs().groupby(group).mean()
    return np.log(maes.map(lambda x: max(x, floor))).mean()


def train_model_regression(X, X_test, target_col, params, folds, model_type='lgb', eval_metric='mae', columns=None,
                           model=None, verbose=1000, early_stopping_rounds=200, n_estimators=50000, metrics_dict=None,
                           beta_encoding=False, cat_col=None, encode_col=None, N_min=1, encoding_nan=False, encoding_type='mean',
                           feature_importance=True):
    """
    A function to train a variety of regression models.
    Returns dictionary with oof predictions, test predictions, scores and, if necessary, feature importances.

    :param: X - training data, can be pd.DataFrame (after normalizing)
    :param: X_test - test data, can be pd.DataFrame (after normalizing)
    :param: target_col - target col name
    :param: folds - folds to split data
    :param: model_type - type of model to use
    :param: eval_metric - metric to use
    :param: columns - columns to use. If None - use all columns
    :param: plot_feature_importance - whether to plot feature importance of LGB
    :param: model - sklearn model, works only for "sklearn" model type
    :param: beta_encoding - do beta_encoding in k-folds
    :param: feature_importance - return feature importance
    :param: encode_col - the columns used for encoding
    """
    if beta_encoding:
        if (not isinstance(encode_col, list)) and (not isinstance(encode_col, np.ndarray)):
            raise TypeError('cat_col should be list or np.ndarry')

    if columns is None:
        columns = [col for col in X.columns if col != target_col]

    if feature_importance:
        if model_type == 'sklearn':
            feature_importance = False

    if metrics_dict is None:
        metrics_dict = {
            'mae': {
                'lgb_metric_name': 'mae',
                'catboost_metric_name': 'MAE',
                'xgb_metric_name': 'mae',
                'sklearn_scoring_function': metrics.mean_absolute_error
            },
            'group_mae': {
                'lgb_metric_name': 'mae',
                'catboost_metric_name': 'MAE',
                'xgb_metric_name': 'mae',
                'scoring_function': group_mean_log_mae
            },
            'mse': {
                'lgb_metric_name': 'mse',
                'catboost_metric_name': 'MSE',
                'xgb_metric_name': 'rmse',
                'sklearn_scoring_function': metrics.mean_squared_error
            }
        }

    result_dict = {}
    # list of scores on folds
    models = []
    scores = []  # list of scores on folds
    oof = np.zeros(len(X))  # out-of-fold predictions on train data
    prediction = np.zeros(len(X_test))  # averaged predictions on train data
    feature_importance_df = pd.DataFrame()
    for fold_n, (train_index, valid_index) in enumerate(folds.split(X[columns], X[target_col])):
        print('Fold {} started at {}'.format({fold_n + 1}, {time.ctime()}))
        X_train, y_train, X_valid, y_valid = X[columns].iloc[train_index], X[target_col].iloc[train_index], X[columns].iloc[valid_index],  X[target_col].iloc[valid_index]

        if beta_encoding:
            # encode variables
            feature_col = []
            for var_name in tqdm(encode_col):
                # fit encoder
                be = BetaEncoder(var_name, encoding_nan)
                be.fit(X_train[[var_name, target_col]], target_col)

                feature_name = var_name + encoding_type
                X_train[feature_name] = be.transform(X_train[[var_name]], encoding_type, N_min)
                X_valid[feature_name] = be.transform(X_valid[[var_name]], encoding_type, N_min)
                X_test[feature_name] = be.transform(X_test[[var_name]], encoding_type, N_min)
                feature_col.append(feature_name)
                gc.collect()
            columns += feature_col

        if model_type == 'lgb':
            model = lgb.LGBMRegressor(**params, n_estimators=n_estimators, n_jobs=-1)
            model.fit(X_train, y_train,
                      eval_set=[(X_train, y_train), (X_valid, y_valid)],
                      eval_metric=metrics_dict[eval_metric]['lgb_metric_name'],
                      categorical_feature=cat_col,
                      verbose=verbose,
                      early_stopping_rounds=early_stopping_rounds)

            y_pred_valid = model.predict(X_valid)
            y_pred = model.predict(X_test[columns], num_iteration=model.best_iteration_)
        elif model_type == 'xgb':
            model = xgb.XGBRegressor(**params, n_estimators=n_estimators, n_jobs=-1)
            model.fit(X_train, y_train,
                      eval_set=[(X_train, y_train), (X_valid, y_valid)],
                      eval_metric=metrics_dict[eval_metric]['xgb_metric_name'],
                      verbose=verbose,
                      early_stopping_rounds=early_stopping_rounds,
                      )

            y_pred_valid = model.predict(X_valid, ntree_limit=model.best_ntree_limit)
            y_pred = model.predict(X_test[columns], ntree_limit=model.best_ntree_limit)
        elif model_type == 'cat':
            model = CatBoostRegressor(iterations=n_estimators,
                                      eval_metric=metrics_dict[eval_metric]['catboost_metric_name'],
                                      loss_function=metrics_dict[eval_metric]['catboost_metric_name'],
                                      verbose=verbose,
                                      cat_features=cat_col,
                                      **params)
            model.fit(X_train, y_train, eval_set=(X_valid, y_valid))
            gc.collect()

            y_pred_valid = model.predict(X_valid)
            y_pred = model.predict(X_test[columns])
        elif model_type == 'sklearn':
            model = model
            model.fit(X_train, y_train)

            y_pred_valid = model.predict(X_valid)
            score = metrics_dict[eval_metric]['sklearn_scoring_function'](y_valid, y_pred_valid)
            print("Fold {}. {}: {}.".format({fold_n}, {eval_metric}, {score}))
            print('')

            y_pred = model.predict(X_test[columns])

        oof[valid_index] = y_pred_valid.reshape(-1,)
        if eval_metric != 'group_mae':
            scores.append(metrics_dict[eval_metric]['sklearn_scoring_function'](y_valid, y_pred_valid))
        else:
            scores.append(metrics_dict[eval_metric]['scoring_function'](y_valid, y_pred_valid,
                                                                        X_valid['group']))
        prediction += y_pred.reshape(-1,)

        if feature_importance:
            fold_importance_df = pd.DataFrame()
            fold_importance_df['feature'] = columns
            fold_importance_df['importance'] = model.feature_importances_
            fold_importance_df['fold'] = fold_n + 1
            feature_importance_df = pd.concat([feature_importance_df, fold_importance_df], axis=0)

        gc.collect()

    prediction /= folds.n_splits
    print('CV mean score: {0:.4f}, std: {1:.4f}.'.format(np.mean(scores), np.std(scores)))

    result_dict['oof'] = oof
    result_dict['prediction'] = prediction
    result_dict['scores'] = scores

    return result_dict, feature_importance_df


def train_model_classification(X, X_test, target_col, params, folds, model_type='lgb', eval_metric='auc', columns=None,
                           model=None, verbose=1000, early_stopping_rounds=200, n_estimators=50000, metrics_dict=None,
                           beta_encoding=False, cat_col=None, encode_col=None, N_min=1, encoding_nan=False, encoding_type='mean',
                           feature_importance=False):
    """
    A function to train a variety of regression models.
    Returns dictionary with oof predictions, test predictions, scores and, if necessary, feature importances.

    :param: X - training data, can be pd.DataFrame (after normalizing)
    :param: X_test - test data, can be pd.DataFrame (after normalizing)
    :param: target_col - target col name
    :param: folds - folds to split data
    :param: model_type - type of model to use
    :param: eval_metric - metric to use
    :param: columns - columns to use. If None - use all columns
    :param: plot_feature_importance - whether to plot feature importance of LGB
    :param: model - sklearn model, works only for "sklearn" model type
    :param: beta_encoding - do beta_encoding in k-folds
    :param: feature_importance - return feature importance
    :param: encode_col - the columns used for encoding
    """
    if beta_encoding:
        if (not isinstance(encode_col, list)) and (not isinstance(encode_col, np.ndarray)):
            raise TypeError('cat_col should be list or np.ndarry')

    if columns is None:
        columns = [col for col in X.columns if col != target_col]

    if feature_importance:
        if model_type == 'sklearn':
            feature_importance = False

    if metrics_dict is None:
        metrics_dict = {
            'auc': {
                'lgb_metric_name': eval_auc,
                'catboost_metric_name': 'AUC',
                'xgb_metric_name': 'auc',
                'sklearn_scoring_function': metrics.roc_auc_score
            },
        }

    X = X[columns]
    X_test = X_test[columns]

    result_dict = {}
    # list of scores on folds
    scores = []  # list of scores on folds
    oof = np.zeros((len(X), 1))  # out-of-fold predictions on train data
    prediction = np.zeros((len(X_test), 1))  # averaged predictions on train data
    feature_importance_df = pd.DataFrame()
    for fold_n, (train_index, valid_index) in enumerate(folds.split(X[columns], X[target_col])):
        print('Fold {} started at {}'.format({fold_n + 1}, {time.ctime()}))
        X_train, y_train, X_valid, y_valid = X[columns].iloc[train_index], X[target_col].iloc[train_index], X[columns].iloc[valid_index],  X[target_col].iloc[valid_index]

        if beta_encoding:
            # encode variables
            feature_col = []
            for var_name in tqdm(encode_col):
                # fit encoder
                be = BetaEncoder(var_name, encoding_nan)
                be.fit(X_train[[var_name, target_col]], target_col)

                feature_name = var_name + encoding_type
                X_train[feature_name] = be.transform(X_train[[var_name]], encoding_type, N_min)
                X_valid[feature_name] = be.transform(X_valid[[var_name]], encoding_type, N_min)
                X_test[feature_name] = be.transform(X_test[[var_name]], encoding_type, N_min)
                feature_col.append(feature_name)
                gc.collect()
            columns += feature_col

        if model_type == 'lgb':
            model = lgb.LGBMClassifier(**params, n_estimators=n_estimators, n_jobs=-1)
            model.fit(X_train, y_train,
                      eval_set=[(X_train, y_train), (X_valid, y_valid)],
                      eval_metric=metrics_dict[eval_metric]['lgb_metric_name'],
                      categorical_feature=cat_col,
                      verbose=verbose,
                      early_stopping_rounds=early_stopping_rounds)

            y_pred_valid = model.predict_proba(X_valid)[:, 1]
            y_pred = model.predict_proba(X_test[columns], num_iteration=model.best_iteration_)[:, 1]
        elif model_type == 'xgb':
            model = xgb.XGBClassifier(**params, n_estimators=n_estimators, n_jobs=-1)
            model.fit(X_train, y_train,
                      eval_set=[(X_train, y_train), (X_valid, y_valid)],
                      eval_metric=metrics_dict[eval_metric]['xgb_metric_name'],
                      verbose=verbose,
                      early_stopping_rounds=early_stopping_rounds,
                      )

            y_pred_valid = model.predict_proba(X_valid, ntree_limit=model.best_ntree_limit)[:, 1]
            y_pred = model.predict_proba(X_test[columns], ntree_limit=model.best_ntree_limit)[:, 1]
        elif model_type == 'cat':
            model = CatBoostClassifier(iterations=n_estimators,
                                       eval_metric=metrics_dict[eval_metric]['catboost_metric_name'],
                                       loss_function=metrics_dict[eval_metric]['catboost_metric_name'],
                                       verbose=verbose,
                                       cat_features=cat_col,
                                       **params)
            model.fit(X_train, y_train,, eval_set=(X_valid, y_valid))
            gc.collect()

            y_pred_valid = model.predict_proba(X_valid)[:, 1]
            y_pred = model.predict_proba(X_test[columns])[:, 1]
        elif model_type == 'sklearn':
            model = model
            model.fit(X_train, y_train)

            y_pred_valid = model.predict_proba(X_valid)[:, 1]
            score = metrics_dict[eval_metric]['sklearn_scoring_function'](y_valid, y_pred_valid)
            print("Fold {}. {}: {}.".format({fold_n}, {eval_metric}, {score}))
            print('')

            y_pred = model.predict_proba(X_test[columns])[:, 1]

        oof[valid_index] = y_pred_valid.reshape(-1, 1)
        if eval_metric != 'group_mae':
            scores.append(metrics_dict[eval_metric]['sklearn_scoring_function'](y_valid], y_pred_valid))
        else:
            scores.append(metrics_dict[eval_metric]['scoring_function'](y_valid, y_pred_valid,
                                                                        X_valid['group']))
        prediction += y_pred.reshape(-1, 1)
        if feature_importance:
            fold_importance_df = pd.DataFrame()
            fold_importance_df['feature'] = columns
            fold_importance_df['importance'] = model.feature_importances_
            fold_importance_df['fold'] = fold_n + 1
            feature_importance_df = pd.concat([feature_importance_df, fold_importance_df], axis=0)
        gc.collect()

    prediction /= folds.n_splits
    print('CV mean score: {0:.4f}, std: {1:.4f}.'.format(np.mean(scores), np.std(scores)))

    result_dict['oof'] = oof
    result_dict['prediction'] = prediction
    result_dict['scores'] = scores

    return result_dict, feature_importance_df



"""
Authors:
        ITryagain <long452a@163.com>

Reference:
        https://www.kaggle.com/vprokopev/mean-likelihood-encodings-a-comprehensive-study
        https://www.kaggle.com/c/avito-demand-prediction/discussion/60059 (Bayesian target encoding)
        https://www.kaggle.com/tnarik/likelihood-encoding-of-categorical-features (likelihood encoding)
        https://zhuanlan.zhihu.com/p/40231966
        https://mp.weixin.qq.com/s/U93vvFwZ8vSJuswk24yc6w

"""
import numpy as np
import pandas as pd
from tqdm import tqdm
from sklearn.model_selection import StratifiedKFold


def one_hot_encode(train_data, test_data, columns):
    """
    :return: Returns the train and test DataFrame with encoded columns
    """
    all_data = pd.concat([train_data, test_data], axis=0)
    encoded_cols = []
    with tqdm(range(len(columns)), 'one_hot_encoding...') as t:
        for i in t:
            try:
                col = columns[i]
                encoded_cols.append(pd.get_dummies(all_data[col], prefix='one_hot_'+col))
            except StopIteration:
                break
    all_encoded = pd.concat(encoded_cols, axis=1)
    return all_encoded.iloc[:train_data.shape[0], :], all_encoded.iloc[train_data.shape[0]:, :]


def label_encode(train_data, test_data, columns, is_concat=False, fill_na=True, sort=True):
    """
    :param: fill_na: if fill_na == True will fill nan with -1, default True
    :param: is_concat: encoding时是否合并train_data和test_data
    :param: sort: 映射时是否需要排序
    :return: Returns the train and test DataFrame with encoded columns
    """
    encoded_cols = []
    if is_concat:
        all_data = pd.concat([train_data, test_data], axis=0)
        with tqdm(range(len(columns)), 'label_encoding...') as t:
            for i in t:
                try:
                    col = columns[i]
                    factorised = pd.factorize(all_data[col], sort=sort)[1]
                    labels = pd.Series(range(len(factorised)), index=factorised)
                    encoded_col = all_data[col].map(labels)
                    if fill_na:
                        encoded_col[encoded_col.isnull()] = -1
                    encoded_cols.append(pd.DataFrame({'label_'+col: encoded_col}))
                except StopIteration:
                    break
    else:
        with tqdm(range(len(columns)), 'label_encoding...') as t:
            for i in t:
                try:
                    col = columns[i]
                    factorised = pd.factorize(train_data[col], sort=sort)[1]
                    labels = pd.Series(range(len(factorised)), index=factorised)
                    encoded_train = train_data[col].map(labels)
                    encode_test = test_data[col].map(labels)
                    encoded_col = pd.concat([encoded_train, encode_test], axis=0)
                    if fill_na:
                        encoded_col[encoded_col.isnull()] = -1
                    encoded_cols.append(pd.DataFrame({'label_'+col: encoded_col}))
                except StopIteration:
                    break
    all_encoded = pd.concat(encoded_cols, axis=1)
    return all_encoded.iloc[:train_data.shape[0], :], all_encoded.iloc[train_data.shape[0]:, :]


def freq_encode(train_data, test_data, columns, is_concat=False, fill_na=True):
    """
    :param: fill_na: if fill_na == True will fill nan with 0, default True
    :param: is_concat: encoding时是否合并train_data和test_data
    :return: Returns the train and test DataFrame with encoded columns
    """
    encoded_cols = []
    nsamples = train_data.shape[0]
    if is_concat:
        all_data = pd.concat([train_data, test_data], axis=0)
        with tqdm(range(len(columns)), 'freq_encoding...') as t:
            for i in t:
                try:
                    col = columns[i]
                    freqs_cat = all_data.groupby(col)[col].count() / nsamples
                    encoded_col = all_data[col].map(freqs_cat)
                    if fill_na:
                        encoded_col[encoded_col.isnull()] = 0
                    encoded_cols.append(pd.DataFrame({'freq_'+col: encoded_col}))
                except StopIteration:
                    break
    else:
        with tqdm(range(len(columns)), 'freq_encoding...') as t:
            for i in t:
                try:
                    col = columns[i]
                    freqs_cat = train_data.groupby(col)[col].count() / nsamples
                    encoded_train = train_data[col].map(freqs_cat)
                    encoded_test = test_data[col].map(freqs_cat)
                    encoded_col = pd.concat([encoded_train, encoded_test], axis=0)
                    if fill_na:
                        encoded_col[encoded_col.isnull()] = 0
                    encoded_cols.append(pd.DataFrame({'freq_'+col: encoded_col}))
                except StopIteration:
                    break
    all_encoded = pd.concat(encoded_cols, axis=1)
    return all_encoded.iloc[:train_data.shape[0], :], all_encoded.iloc[train_data.shape[0]:, :]


def mean_encode(train_data, test_data, columns, target_col, reg_method=None,
                alpha=5, add_random=False, rmean=0, rstd=0.1, folds=5, seed=2019):
    """
    :return: Returns the train and test DataFrame with encoded columns
    """
    encoded_cols = []
    target_mean_global = train_data[target_col].mean()
    with tqdm(range(len(columns))) as t:
        for i in t:
            try:
                col = columns[i]
                # Getting means for test data
                nrows_cat = train_data.groupby(col)[target_col].count()
                target_means_cats = train_data.groupby(col)[target_col].mean()
                target_means_cats_adj = (target_means_cats * nrows_cat +
                                         target_mean_global * alpha) / (nrows_cat + alpha)
                # Mapping means to test data
                encoded_test = test_data[col].map(target_means_cats_adj)
                # Getting a train encodings
                if reg_method == 'expanding_mean':
                    train_data_shuffled = train_data.sample(frac=1, random_state=1)
                    cumsum = train_data_shuffled.groupby(col)[target_col].cumsum() - train_data_shuffled[target_col]
                    cumcnt = train_data_shuffled.groupby(col).cumcount()
                    encoded_train = cumsum / cumcnt
                    encoded_train.fillna(target_mean_global, inplace=True)
                    if add_random:
                        encoded_train = encoded_train + np.random.normal(loc=rmean, scale=rstd,
                                                                         size=(encoded_train.shape[0]))
                elif (reg_method == 'k_fold') and (folds > 1):
                    kfold = StratifiedKFold(n_splits=folds, shuffle=True, random_state=seed)
                    parts = []
                    for tr_in, val_ind in enumerate(kfold.split(train_data.values, train_data.values)):
                        # divide data
                        df_for_estimation, df_estimated = train_data.iloc[tr_in], train_data.iloc[val_ind]
                        # getting means on data for estimation (all folds except estimated)
                        nrows_cat = df_for_estimation.groupby(col)[target_col].count()
                        target_means_cats = df_for_estimation.groupby(col)[target_col].mean()
                        target_means_cats_adj = (target_means_cats * nrows_cat +
                                                 target_mean_global * alpha) / (nrows_cat + alpha)
                        # Mapping means to estimated fold
                        encoded_train_part = df_estimated[col].map(target_means_cats_adj)
                        if add_random:
                            encoded_train_part = encoded_train_part + np.random.normal(loc=rmean, scale=rstd,
                                                                                       size=(
                                                                                       encoded_train_part.shape[0]))
                        # Saving estimated encodings for a fold
                        parts.append(encoded_train_part)
                    encoded_train = pd.concat(parts, axis=0)
                    encoded_train.fillna(target_mean_global, inplace=True)
                else:
                    encoded_train = train_data[col].map(target_means_cats_adj)
                    if add_random:
                        encoded_train = encoded_train + np.random.normal(loc=rmean, scale=rstd,
                                                                         size=(encoded_train.shape[0]))

                # Saving the column with means
                encoded_col = pd.concat([encoded_train, encoded_test], axis=0)
                encoded_col[encoded_col.isnull()] = target_mean_global
                encoded_cols.append(pd.DataFrame({'mean_' + target_col + '_' + col: encoded_col}))
            except StopIteration:
                break
    all_encoded = pd.concat(encoded_cols, axis=1)
    return all_encoded.iloc[:train_data.shape[0], :], all_encoded.iloc[train_data.shape[0]:, :]


def bayesian_target_encoding(train_data, valid_data, test_data, columns, target_col, N_min,
                             stat_type='mean', encode_na=False):
    """
    :param target_col: label
    :param N_min: regularization term, N_min 越大，regularization效果越强。
    :param stat_type: 包括 "mean" "mode" "median" "var" "skewness" "kurtosis"
    :param encode_na: 是否对nan进行编码
    :return:Returns the train and test DataFrame with encoded columns
    """
    if not encode_na:
        for col in columns:
            if train_data[col].isna().sum() and valid_data[col].isna().sum() and test_data[col].isna().sum():
                raise ValueError("the col {} contains nan".format({col}))
    else:
        train_data[columns] = train_data[columns].astype(str)
        valid_data[columns] = valid_data[columns].astype(str)
        test_data[columns] = test_data[columns].astype(str)
    feature_cols = []
    prior_mean = np.mean(train_data[target_col])
    with tqdm(range(len(columns))) as t:
        for i in t:
            try:
                col = columns[i]
                new_col = col + '_' + stat_type
                feature_cols.append(new_col)
                stats = train_data[[col, target_col]].groupby(col).agg(['sum', 'count'])[target_col].reset_index()
                train_data[new_col] = _bayesian_target_encoding_trans(train_data[[col]], stat_type,
                                                                                stats, prior_mean, N_min)
                valid_data[new_col] = _bayesian_target_encoding_trans(valid_data[[col]], stat_type,
                                                                                stats, prior_mean, N_min)
                test_data[new_col] = _bayesian_target_encoding_trans(test_data[[col]], stat_type,
                                                                                stats, prior_mean, N_min)
            except StopIteration:
                break
    if encode_na:
        train_data[columns] = train_data[columns].astype('category')
        valid_data[columns] = valid_data[columns].astype('category')
        test_data[columns] = test_data[columns].astype('category')
    return train_data, valid_data, test_data, feature_cols


def _bayesian_target_encoding_trans(data, stat_type, stats, prior_mean, N_min):
    df_stats = pd.merge(data, stats, how='left')
    df_stats['sum'].fillna(value=prior_mean, inplace=True)
    df_stats['count'].fillna(value=1.0, inplace=True)
    N_prior = np.maximum(N_min - df_stats['count'].values, 0)
    alpha_prior = prior_mean * N_prior
    beta_prior = (1 - prior_mean) * N_prior

    alpha = alpha_prior + df_stats['sum']
    beta = beta_prior + df_stats['count'] - df_stats['sum']

    if stat_type == 'mean':
        num = alpha
        dem = alpha + beta
    elif stat_type == 'mode':
        num = alpha - 1
        dem = alpha + beta - 2
    elif stat_type == 'median':
        num = alpha - 1 / 3
        dem = alpha + beta - 2 / 3
    elif stat_type == 'var':
        num = alpha * beta
        dem = (alpha + beta) ** 2 * (alpha + beta + 1)
    elif stat_type == 'skewness':
        num = 2 * (beta - alpha) * np.sqrt(alpha + beta + 1)
        dem = (alpha + beta + 2) * np.sqrt(alpha * beta)
    elif stat_type == 'kurtosis':
        num = 6 * (alpha - beta) ** 2 * (alpha + beta + 1) - alpha * beta * (alpha + beta + 2)
        dem = alpha * beta * (alpha + beta + 2) * (alpha + beta + 3)
    else:
        num = prior_mean
        dem = np.ones_like(N_prior)
    value = num / dem
    value[np.isnan(value)] = np.nanmedian(value)
    return value


def nan_encoding(train_data, test_data, columns):
    """
    :return: Returns the train and test DataFrame with encoded columns
    """
    all_data = pd.concat([train_data, test_data], axis=0)
    encoded_cols = []
    with tqdm(range(len(columns)), 'nan_encoding...') as t:
        for i in t:
            try:
                col = columns[i]
                encoded_cols.append(pd.get_dummies(all_data[col], prefix='nan_' + col, drop_first=True,
                                                   dummy_na=True))
            except StopIteration:
                break
    all_encoded = pd.concat(encoded_cols, axis=1)
    return all_encoded.iloc[:train_data.shape[0], :], all_encoded.iloc[train_data.shape[0]:, :]


def count_encoding(train_data, test_data, columns, is_concat=False, fill_na=True):
    """
    :param: fill_na: if fill_na == True will fill nan with 0, default True
    :param: is_concat: encoding时是否合并train_data和test_data
    :return: Returns the train and test DataFrame with encoded columns
    """
    encoded_cols = []
    if is_concat:
        all_data = pd.concat([train_data, test_data], axis=0)
        with tqdm(range(len(columns)), 'count_encoding...') as t:
            for i in t:
                try:
                    col = columns[i]
                    count_cat = all_data.groupby(col)[col].count()
                    encoded_col = all_data[col].map(count_cat)
                    if fill_na:
                        encoded_col[encoded_col.isnull()] = 0
                    encoded_cols.append(pd.DataFrame({'count_' + col: encoded_col}))
                except StopIteration:
                    break
    else:
        with tqdm(range(len(columns)), 'freq_encoding...') as t:
            for i in t:
                try:
                    col = columns[i]
                    count_cat = train_data.groupby(col)[col].count()
                    encoded_train = train_data[col].map(count_cat)
                    encoded_test = test_data[col].map(count_cat)
                    encoded_col = pd.concat([encoded_train, encoded_test], axis=0)
                    if fill_na:
                        encoded_col[encoded_col.isnull()] = 0
                    encoded_cols.append(pd.DataFrame({'count_' + col: encoded_col}))
                except StopIteration:
                    break
    all_encoded = pd.concat(encoded_cols, axis=1)
    return all_encoded.iloc[:train_data.shape[0], :], all_encoded.iloc[train_data.shape[0]:, :]


class BetaEncoder(object):

    def __init__(self, group, encoding_nan=False):

        self.group = group
        self.stats = None
        self.encoding_nan = encoding_nan

    # get counts from df
    def fit(self, df, target_col):
        self.prior_mean = np.mean(df[target_col])
        tmp = df[[target_col, self.group]].copy()
        if self.encoding_nan:
            tmp[self.group] = tmp[self.group].astype(str)
        stats = tmp[[target_col, self.group]].groupby(self.group)
        stats = stats.agg(['sum', 'count'])[target_col]
        stats.rename(columns={'sum': 'n', 'count': 'N'}, inplace=True)
        stats.reset_index(level=0, inplace=True)
        self.stats = stats

    # extract posterior statistics
    def transform(self, df, stat_type, N_min=1):
        tmp = df[[self.group]].copy()
        if self.encoding_nan:
            tmp[self.group] = tmp[self.group].astype(str)
        df_stats = pd.merge(tmp[[self.group]], self.stats, how='left')
        n = df_stats['n'].copy()
        N = df_stats['N'].copy()

        # fill in missing
        nan_indexs = np.isnan(n)
        n[nan_indexs] = self.prior_mean
        N[nan_indexs] = 1.0

        # prior parameters
        N_prior = np.maximum(N_min - N, 0)
        alpha_prior = self.prior_mean * N_prior
        beta_prior = (1 - self.prior_mean) * N_prior

        # posterior parameters
        alpha = alpha_prior + n
        beta = beta_prior + N - n

        # calculate statistics
        if stat_type == 'mean':
            num = alpha
            dem = alpha + beta

        elif stat_type == 'mode':
            num = alpha - 1
            dem = alpha + beta - 2

        elif stat_type == 'median':
            num = alpha - 1 / 3
            dem = alpha + beta - 2 / 3

        elif stat_type == 'var':
            num = alpha * beta
            dem = (alpha + beta) ** 2 * (alpha + beta + 1)

        elif stat_type == 'skewness':
            num = 2 * (beta - alpha) * np.sqrt(alpha + beta + 1)
            dem = (alpha + beta + 2) * np.sqrt(alpha * beta)

        elif stat_type == 'kurtosis':
            num = 6 * (alpha - beta) ** 2 * (alpha + beta + 1) - alpha * beta * (alpha + beta + 2)
            dem = alpha * beta * (alpha + beta + 2) * (alpha + beta + 3)

        else:
            num = self.prior_mean
            dem = np.ones_like(N_prior)

        # replace missing
        value = num / dem
        value[np.isnan(value)] = np.nanmedian(value)
        return value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 15 19:44:03 2019

@author: soushigou
"""

import pandas as pd

def load_data(str):
    try:
        data=pd.read_csv("../dataset/"+str)
    except:
        traceback.print_exc()
        
    return data



from sklearn import preprocessing

# 均匀分桶：主要用于将连续特征转化为类别离散特征
# 先使得特征均匀
def norm_data(df,features):   
    scaler = preprocessing.QuantileTransformer(random_state=0)
    scaler.fit(df[features].values.reshape(-1, 1)) 
    new_c = 'cat_'+features
    df[new_c]=scaler.transform(df[features].values.reshape(-1, 1))
    return df
#norm_split(train, test, 'fuzz_qratio')
#norm_data(data, 'fuzz_qratio')

# 再对均匀分布的数据分桶
# the continous_to_category function will create new category features 
def continous_to_category(data, features):
    data = norm_data(data, features)
    bins_list = []
    new_c = 'cat_'+features
    bins_num = 10
    i_min = data[new_c].min()-0.001
    i_max = data[new_c].max()+0.001
    for i in range(bins_num+1):
        bins_list.append(i_min+(i_max-i_min)/10*i)
    data[new_c] = pd.cut(data[new_c], bins=bins_list, labels=np.arange(bins_num))
    return data
# 用法如下：
# newdata = continous_to_category(data, 'feature_name')
# 之后newdata中会出现新特征"cat_feature_name"


import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import numpy as np
import math
from scipy import stats


class Overview(object):
    @staticmethod
    def check_missing_data(data):
        """
        检查数据中缺失值

        :param: data:输入数据
        :resturn :缺失值信息表格
        """
        total = data.isnull().sum()
        percent = (data.isnull().sum() / data.isnull().count() * 100)
        res = pd.concat([total, percent], axis=1, keys=['Total', 'Percent'])
        types = []
        for col in data.columns:
            dtype = str(data[col].dtype)
            types.append(dtype)
        res['Types'] = types
        return res

    @staticmethod
    def plot_feature_distribution(df, y_features, x_feature=None, hue=None, columns=5):
        """
        画m×n的折线图/密度图
        :param: df:数据
        :param: y_features:纵坐标特征列表（list，长度小于等于100）
        :param: x_feature:横坐标的特征（str）
        :param: hue:分类目标特征 (str)
        :param: columns:每一行绘制的图表数量

        　　当传入的df为一个pd.DataFrame()对象时，每一个图表仅绘制该对象的数据信息；当df为一个字典时(
        字典的键应为其对应的字典中的数据的标签，如：{'train':df_train,'test':df_test})，每个图表绘
        制字典中存储的所有的数据信息。
        　　当x_feature为空时，绘制y_features中每一个feature的密度图；否则绘制以x_feature为横轴的
        折线图。
        """
        # 检查参数 check params
        if hue and (x_feature is None):
            raise Exception("Wrong Params!")

        n = columns
        hight = math.ceil(len(y_features) / n)

        width = 36
        height = width / n * 2 / 3

        i = 0
        for feature in y_features:
            i = (i + 1) % n
            if i == 1:
                fig, ax = plt.subplots(1, n, figsize=(width, height))
                plt.subplot(1, n, 1)
            elif i == 0:
                plt.subplot(1, n, n)
            else:
                plt.subplot(1, n, i)
            if isinstance(df, dict):
                for df_ in df.values():
                    for _ in df.keys():
                        if df[_] is df_: label = _
                    df_ = df_[~df_[feature].isna()]

                    if x_feature:
                        df_ = df_[~df_[x_feature].isna()]
                        if hue:
                            sns.lineplot(x=x_feature, y=feature, hue=hue, data=df_, label=label)
                        else:
                            sns.lineplot(x=x_feature, y=feature, data=df_, label=label)
                    else:
                        sns.distplot(df_[feature], label=label)
                plt.legend()
                if x_feature:
                    plt.xlabel(x_feature, fontsize=14)
                else:
                    plt.xlabel(feature, fontsize=14)
            else:
                df = df[~df[feature].isna()]
                if x_feature:
                    df = df[~df[x_feature].isna()]
                    if hue:
                        sns.lineplot(x=x_feature, y=feature, hue=hue, data=df)
                    else:
                        sns.lineplot(x=x_feature, y=feature, data=df)
                    plt.xlabel(x_feature, fontsize=14)
                else:
                    sns.distplot(df[feature])
                    plt.xlabel(feature, fontsize=14)

            plt.tick_params(axis='x', which='major', labelsize=14, pad=0)
            plt.tick_params(axis='y', which='major', labelsize=14)
            if i == 0:
                plt.show()

    @staticmethod
    def describe_table(df):
        print("Dataset Shape: {}".format({df.shape}))
        summary = pd.DataFrame(df.dtypes, columns=['dtypes']).reset_index()
        summary['Name'] = summary['index']
        summary = summary[['Name', 'dtypes']]
        summary['Missing'] = df.isnull().sum().values
        summary['Uniques'] = df.nunique().values
        summary['First Value'] = df.loc[0].values
        summary['Second Value'] = df.loc[1].values
        summary['Third Value'] = df.loc[2].values

        for name in summary['Name'].value_counts().index:
            summary.loc[summary['Name'] == name, 'Entropy'] = round(stats.entropy(df[name].value_counts(normalize=True),
                                                                                  base=2), 2)
        return summary

    @staticmethod
    def feature_quantiles(df, columns):
        print("Features Quantitles:")
        print(df[columns].quantile([0.01, .025, .1, .25, .5, .75, .975, .99]))

    @staticmethod
    def cal_outliers(df_num):

        # calculating mean and std of the array
        data_mean, data_std = np.mean(df_num), np.std(df_num)

        # setting the cut line to both higher and lower values
        # this value can change
        cut = data_mean * 3

        # calculating the higher and lower cut values
        lower, upper = data_mean - cut, data_mean + cut

        # creating an array of lower, higher and total outlier values
        outliers_lower = [x for x in df_num if x < lower]
        outliers_higher = [x for x in df_num if x > upper]
        outliers_total = outliers_lower + outliers_higher

        # array without outlier values
        outliers_removed = [x for x in df_num if lower < x < upper]

        print('Identified lowest outliers: {}'.format(len(outliers_lower)))
        print('Identified upper outliers: {}'.format(len(outliers_higher)))
        print('Total outlier observations: {}'.format(len(outliers_total)))
        print('Non-Outlier observations: {}'.format(len(outliers_removed)))
        print("Total percent of Outliers: ", round((len(outliers_total) / len(outliers_removed)) * 100, 4))

        return



from .Encoding import one_hot_encode, label_encode, freq_encode, mean_encode, \
    bayesian_target_encoding, nan_encoding, count_encoding, BetaEncoder

__all__ = ['one_hot_encode',
           'label_encode',
           'freq_encode',
           'mean_encode',
           'bayesian_target_encoding',
           'nan_encoding',
           'count_encoding',
           'BetaEncoder']



"""
Authors:
        Zwp <570972004@qq.com>

Referance:
    https://github.com/fmfn/BayesianOptimization

introduce:
        使用贝叶斯优化调整参数，是对bayes_opt的高层封装

"""
from bayes_opt import BayesianOptimization
import numpy as np
import pandas as pd

class ParamOptimizer(object):
    def __init__(self, score_func, score_func_parmas):
        """
        :param: score_func: 计算得分的函数 handyML.models.model.train_model_regression / handyML.models.model.train_model_regression
        :param: score_func_parmas: 模型函数的参数
        """
        self.score_func = score_func
        self.score_func_parmas = score_func_parmas
        
        self.solid_params = {} #恒定的训练参数
        self.optimizer = None
        self.int_params = None
    
    def optimize(self, params, int_params=None):
        """
        :param: params: 参数字典  {'num_leaves':1, 'max_depth':(1,2)}
        :param: int_params: 参数字典中整型参数列表
        """
        self.int_params = int_params
        opt_params = {}
        for k in params.keys():
            if isinstance(params[k], tuple) or isinstance(params[k], list):
                opt_params[k] = params[k]
            else:
                self.solid_params[k] = params[k]
        self.optimizer = BayesianOptimization(self._optimize_func, opt_params, random_state=11)
        self.optimizer.maximize()
        
        print("Best score:",self.optimizer.max['target'])
        return self.optimizer.max['params']
    
    def _optimize_func(self,**params):
        """
        内部优化函数，返回当前参数下的local CV
        """
        p = self.solid_params.copy()
        for k in params.keys():
            if k in self.int_params:
                p[k] = int(round(params[k]))
            else:
                p[k] = params[k]
        print(p['min_data_in_leaf'])
        self.score_func_parmas['params'] = p
        result_dict, _ = self.score_func(**self.score_func_parmas)
        return np.mean(result_dict['scores'])






from setuptools import find_packages
from setuptools import setup
from os import path as os_path

this_directory = os_path.abspath(os_path.dirname(__file__))

def read_file(filename):
    with open(os_path.join(this_directory, filename), encoding='utf-8') as f:
        long_description = f.read()
    return long_description

def read_requirement(filename):
    return [line.strip() for line in read_file(filename).splitlines() if not line.startswith('#')]

setup(
    name="handyML",
    version="0.0.1b",
    author="ITryagain",
    author_email="long452a@163.com",
    python_requires='>=3.6.0',
    description="A library for data science",
    long_description=read_file('README.md'),
    long_description_content_type="text/markdown",
    url="https://github.com/leo6033/handyML",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    install_requires=read_requirement('requirements.txt'),
)